from pygel3d import hmesh


def find_distance_2(hull, points):
    # Construct PyGEL Manifold from the convex hull
    m = hmesh.Manifold()
    for s in hull.simplices:
        m.add_face(hull.points[s])

    dist = hmesh.MeshDistance(m)
    res = []
    for p in points:
        # Get the distance to the point
        # But don't trust its sign, because of possible
        # wrong orientation of mesh face
        d = dist.signed_distance(p)

        # Correct the sign with ray inside test
        if dist.ray_inside_test(p):
            if d > 0:
                d *= -1
        else:
            if d < 0:
                d *= -1

        res.append((d, p))

    return res



from time import perf_counter
from functools import wraps


def logger(debug=False):
    def decorator(f):
        if callable(f):
            @wraps(f)
            def wrapper(*args, **kwargs):
                if debug:
                    print(f'Method "{f.__name__}" was called')
                return f(*args, **kwargs)

            return wrapper

    return decorator


def timer(f):
    if callable(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            start = perf_counter()
            res = f(*args, **kwargs)
            end = perf_counter()
            print(f'Execution time of func "{f.__name__}": {end - start} s')
            return res

        return wrapper

    raise TypeError(f'{f} if not callable.')



# from sys import argv  # for benchmarking only
import logging
import time
# import warnings
from functools import partial
from typing import Union

from MDAnalysis.transformations.wrap import wrap
import MDAnalysis as mda
import numpy as np

from scipy.spatial import ConvexHull
from tqdm import tqdm
# from tqdm import tqdm
from tqdm.contrib.concurrent import process_map
from multiprocessing import Manager

from PUCHIK.grid_project.core.AlphaShape import AlphaShape
from PUCHIK.grid_project.utilities.MoleculeSystem import MoleculeSystem
# Local imports
from PUCHIK.grid_project.utilities.decorators import logger  # , timer
from PUCHIK.grid_project.settings import DEBUG, CPU_COUNT, TQDM_BAR_FORMAT
from PUCHIK.grid_project.core.utils import find_distance, _is_inside

# from .utils_python import find_distance, _is_inside
logging.basicConfig(format='%(message)s')
np.seterr(invalid='ignore', divide='ignore')

"""
    Grid method for analyzing complex shaped structures
"""


class Interface(MoleculeSystem):
    """
        Class creates to create a mesh of points representing different molecule
        types of the system in a grid

        Attributes:
            traj (str): Path to any trajectory format supported by MDAnalysis package
            top (str): Path to any topology format supported by MDAnalysis package. Defaults to None
    """

    def __init__(self, traj, top=None):
        self.grid_matrix = None
        self.u: mda.Universe = mda.Universe(top, traj) if top else mda.Universe(traj)
        self.ag = None
        self.unique_resnames = None
        self.main_structure_selection = None
        self.current_frame = 0

        manager = Manager()
        self._hull = manager.dict()

        self._use_alpha_shape = False

    @property
    def use_alpha_shape(self):
        return self._use_alpha_shape

    @use_alpha_shape.setter
    def use_alpha_shape(self, val):
        if type(val) is not bool:
            raise TypeError('use_alpha_shape must be a boolean')
        self._use_alpha_shape = val

    def select_atoms(self, sel='all'):
        """
        Method for selecting the atoms using MDAnalysis selections

        :param sel: selection string
        :return:
        """
        self.ag = self.u.select_atoms(sel)
        self.unique_resnames = np.unique(self.ag.resnames)
        print('Wrapping trajectory...')
        transform = wrap(self.ag)
        self.u.trajectory.add_transformations(transform)

    def select_structure(self, selection):
        """
        Use this method to select the structure for density calculations․

        :param selection: selection(s) of the main structure
        :return: None
        """
        self.main_structure_selection = selection

    def _get_int_dim(self):
        """
        Utility function to get box dimensions

        Returns:
            Dimensions of the box as an int
        """
        return int(np.ceil(self.u.dimensions[0]))

    @staticmethod
    def make_grid(pbc_dim: int, dim=1, d4=None) -> np.ndarray:
        """
        Returns a 4D matrix

        Args:
             pbc_dim (int): Dimensions of the box
             dim (int): Dimensions of the box
             d4 (int): Returns an 4-D matrix if d4 is given. 4th dimension contains d4 elements
        """

        x = y = z = pbc_dim // dim + 1
        grid_matrix = np.zeros((x, y, z)) if d4 is None else np.zeros((x, y, z, d4))

        return grid_matrix

    @staticmethod
    def check_cube(x: float, y: float, z: float) -> tuple:
        """
        Find to which cube does the atom belong to
        Args:
            x (float): x coordinate
            y (float): y coordinate
            z (float): z coordinate

        Returns:
            tuple: Coordinates of the node inside the grid where the point belongs
        """

        n_x = int(x)
        n_y = int(y)
        n_z = int(z)

        return n_x, n_y, n_z

    @staticmethod
    def make_coordinates(mesh):
        """
        Converts the mesh to coordinates
        Args:
            mesh (np.ndarray):  Mesh to convert into 3D coordinates

        Returns:
            np.ndarray: Ndarray of tuples representing coordinates of each of the points in the mesh
        """

        return np.argwhere(mesh > 0)

    def _calc_mesh(self, grid_dim, selection):
        """
        Calculates the mesh according the atom positions in the box

        Args:
            grid_dim (int): Box dimensions

        Returns:
            np.ndarray: The grid
        """
        atom_group = self.u.select_atoms(selection)
        grid_matrix = self.make_grid(grid_dim)

        for atom in atom_group:
            x, y, z = self.check_cube(*atom.position)
            # Correct outlying coordinates !TODO Create a separate utility function for such checks
            x = min(max(x, 0), grid_dim - 1)
            y = min(max(y, 0), grid_dim - 1)
            z = min(max(z, 0), grid_dim - 1)

            grid_matrix[x, y, z] += 1

        return grid_matrix

    # @logger(DEBUG)
    def calculate_mesh(self, selection, main_structure=False):
        """
        Calculates the mesh using _calc_mesh method
        Args:
            selection (str): Selection for atom group to calculate mesh
            main_structure (bool): use as the main structure if true (e.g. densities are calculated relative to this)
        Returns:
            np.ndarray: Returns the grid matrix
        """
        # define the matrices

        grid_matrix = self._calc_mesh(self._get_int_dim(), selection)  # !TODO _get_int_dim փոխի

        if main_structure:  # if selection is None, then it's the main structure
            self.grid_matrix = grid_matrix

        return grid_matrix

    # @logger(DEBUG)

    def _calculate_density_grid(self, coords, bin_count):
        # Works on a cubic box. !TODO Generalize later
        self.u.trajectory[self.current_frame]  # Set the frame to the current frame. Must be a better way...

        coords = np.array(coords)
        density_grid = np.zeros((bin_count, bin_count, bin_count))

        edges, step = np.linspace(0, self._get_int_dim(), bin_count + 1, retstep=True)
        grid_cell_volume = step ** 3

        # y_edges = np.linspace(0, self._get_int_dim(), bin_count + 1)
        # z_edges = np.linspace(0, self._get_int_dim(), bin_count + 1)

        for x, y, z in coords:
            x_idx = np.digitize(x, edges) - 1
            y_idx = np.digitize(y, edges) - 1
            z_idx = np.digitize(z, edges) - 1

            # This is to ensure indices are within the grid bounds
            x_idx = min(max(x_idx, 0), bin_count - 1)
            y_idx = min(max(y_idx, 0), bin_count - 1)
            z_idx = min(max(z_idx, 0), bin_count - 1)

            density_grid[x_idx, y_idx, z_idx] += 1

        density_grid /= grid_cell_volume

        return density_grid

    def _grid_centers(self, bin_count):
        edges, step = np.linspace(0, self._get_int_dim(), bin_count + 1, retstep=True)
        x_centers = (edges[:-1] + edges[1:]) / 2
        y_centers = (edges[:-1] + edges[1:]) / 2
        z_centers = (edges[:-1] + edges[1:]) / 2
        x_grid, y_grid, z_grid = np.meshgrid(x_centers, y_centers, z_centers, indexing='ij')

        return np.vstack([x_grid.ravel(), y_grid.ravel(), z_grid.ravel()]).T

    def _normalize_density(self, coords, bin_count):
        density_grid = self._calculate_density_grid(coords, bin_count)
        density_grid = density_grid.flatten()

        return density_grid

    def _extract_from_mesh(self, mol_type):
        if mol_type not in self.unique_resnames:
            raise ValueError(
                f'Molecule type "{mol_type}" is not present in the system. Available types: {self.unique_resnames}'
            )

        mol_index = np.where(self.unique_resnames == mol_type)

        return self.grid_matrix[:, :, :, mol_index]

    def _create_hull(self):
        if not self.main_structure_selection:
            raise ValueError('Select the main structure with "select_structure" before running calculations')
        if self._hull.get(self.current_frame):
            # If the hull was calculated for this frame, just return it
            return self._hull[self.current_frame]

        mesh = self.calculate_mesh(selection=self.main_structure_selection, main_structure=True)

        mesh_coords = self.make_coordinates(mesh[:, :, :])
        mesh_coordinates = np.array(mesh_coords)
        # try:
        if self.use_alpha_shape:
            hull = AlphaShape(mesh_coordinates).calculate_as(self.current_frame)
        else:
            hull = ConvexHull(mesh_coordinates)
        self._hull[self.current_frame] = hull
        return hull

    def _calc_dens(self, frame_num, selection, norm_bin_count):
        """
        Calculates the density of selection from interface.

        Args:
            frame_num (int): Number of the frame
            selection (str): Selection of the atom group density of which is to be calculated
            ratio (float): Ratio moltype/water !TODO for testing. Remove later
        Returns:
            tuple: Density array and corresponding distances
        """
        self.current_frame = frame_num
        self.u.trajectory[self.current_frame]

        selection_coords = self.u.select_atoms(selection).positions  # self.make_coordinates(selection_mesh)

        hull = self._create_hull()

        grid_centers = self._grid_centers(bin_count=norm_bin_count)

        distances = np.array(
            find_distance(hull, grid_centers)
        )  # Calculate distances from the interface to each grid cell
        densities = self._normalize_density(
            selection_coords,
            bin_count=norm_bin_count
        )  # Calculate the density of each cell

        indices = np.argsort(distances)
        distances = distances[indices]
        densities = densities[indices]

        return distances, densities

    # @timer
    def calculate_density(self, selection=None, start=0, skip=1, end=None,
                          norm_bin_count=20, cpu_count=CPU_COUNT, mp=True):
        """
        Calculates density of selection from the interface
        :param end: Final frame
        :param norm_bin_count: Bin count for normalization
        :param cpu_count: Number of cores to use
        :param selection: MDAnalysis selection of ag
        :param interface_selection: Selection of what is considered as interface
        :param start: Starting frame
        :param skip: Skip every n-th frame
        :param mp: If False, computation will be performed iteratively on a single process

        :return:
        """
        n_frames = self.u.trajectory.n_frames if end is None else end
        frame_range = range(start, n_frames, skip)
        print(f'Running density calculation for the following atom group: {selection}')
        if mp:
            res = self._mp_calc(self._calc_dens, frame_range, cpu_count, selection=selection,
                                norm_bin_count=norm_bin_count)
        else:
            # Single-process calculation
            start = time.perf_counter()

            res = [None] * len(frame_range)
            for index, i in enumerate(tqdm(frame_range, bar_format=TQDM_BAR_FORMAT)):
                res[index] = self._calc_dens(i, selection, norm_bin_count)
            res = np.array(res)
            print(f'Execution time for {len(frame_range)} frames:', time.perf_counter() - start)

        distances, densities = self._process_result(res)

        # Simply taking the mean might not be the best option
        # distances = distances.mean(axis=0)
        # densities = densities.mean(axis=0)

        return distances, densities

    @staticmethod
    def _process_result(res):
        """ Helper method to correctly calculate the average of the result """
        distances = res[:, 0]
        densities = res[:, 1]
        dim_1, dim_2 = distances.shape
        offset = 50  # an offset to shift the distances to correct positions

        offset_distances = np.zeros((dim_1, dim_2 + offset))
        offset_densities = np.zeros((dim_1, dim_2 + offset))
        for i, arr in enumerate(distances):
            minim = abs(int(arr[0]))

            offset_minim = offset - minim  # Ensure index is non-negative

            offset_distances[i, offset_minim:dim_2 + offset_minim] = arr
            offset_densities[i, offset_minim:dim_2 + offset_minim] = densities[i]

        # Trim zeros
        global_min = abs(int(distances.min()))

        offset_global_min = offset - global_min  # Again, ensure index is non-negative

        final_distances = offset_distances[:, offset_global_min:dim_2 + offset_global_min]
        final_densities = offset_densities[:, offset_global_min:dim_2 + offset_global_min]

        final_distances = final_distances.mean(axis=0, where=final_distances != 0)
        final_densities = final_densities.mean(axis=0, where=final_distances != 0)

        return final_distances, final_densities

    def _calc_count(self, frame_num, selection) -> int:

        self.current_frame = frame_num
        self.u.trajectory[self.current_frame]

        ag = self.u.select_atoms(selection)
        n_atoms = len(ag) // ag.n_residues

        hull = self._create_hull()
        coms = []
        count = 0

        for i in range(ag.n_residues):
            coms.append(ag[i * n_atoms: (i + 1) * n_atoms].center_of_mass())

        for com in coms:
            if _is_inside(com, hull, self.use_alpha_shape):
                count += 1

        return count

    def mol_count(self, selection, start=0, skip=1, end=None, cpu_count=CPU_COUNT):
        """
        The method calculates and returns the number of <selection> molecules inside the interface
        :param selection: Selection for molecules to check if they are solubilized in the NP or no
        :param start:
        :param skip:
        :param end:
        :param cpu_count:
        :return: ndarray containing number of molecules each frame
        """
        n_frames = self.u.trajectory.n_frames if end is None else end
        frame_range = range(start, n_frames, skip)

        print(f'Calculating number of "{selection}" molecules inside the interface')

        res = self._mp_calc(self._calc_count, frame_range, cpu_count, selection=selection)

        return res

    def _calc_hull(self, frame_num):
        self.current_frame = frame_num
        self.u.trajectory[self.current_frame]

        return self._create_hull()

    @logger(DEBUG)
    def calculate_volume(self, area=False, start=0, skip=1, end=None, cpu_count=CPU_COUNT) -> Union[tuple, np.ndarray]:
        """
        Returns the volume of the hull
        :param bool area: If True, return the area of the hull as well
        :param int start: First frame of the trajectory
        :param int skip: How many frames to skip
        :param int end: Final frame of the trajectory
        :param int cpu_count: Sets the number of cores to utilize during the calculation
        :return [tuple, np.ndarray] volume: ndarray containing the volume values of the hull at each frame, or a tuple
        of ndarrays for volumes and areas
        """
        if self.use_alpha_shape:
            raise NotImplementedError('Volume calculations using alpha shapes is not yet implemented')

        n_frames = self.u.trajectory.n_frames if end is None else end
        frame_range = range(start, n_frames, skip)
        print('Calculating the volume of the selected structure')
        hulls = self._mp_calc(self._calc_hull, frame_range, cpu_count)
        return np.array([hull.volume for hull in hulls]) if not area else (
            np.array([hull.volume for hull in hulls]),
            np.array([hull.area for hull in hulls])
        )

    @staticmethod
    def _sp_calc(func):
        ...

    @staticmethod
    def _mp_calc(func, frame_range, cpu_count, **kwargs) -> np.ndarray:
        """
        This method handles multiprocessing
        :param func:
        :param frame_range:
        :param cpu_count:
        :param kwargs:
        :return:
        """
        # start = time.perf_counter()
        per_frame_func = partial(func, **kwargs)
        res = process_map(per_frame_func, frame_range,
                          max_workers=cpu_count,
                          bar_format=TQDM_BAR_FORMAT
                          )
        # print(f'Execution time for {len(frame_range)} frames:', time.perf_counter() - start)

        return np.array(res)


if __name__ == '__main__':
    pass



import numpy as np
from PUCHIK.grid_project.core.utils import _is_inside


def generate_point(dim):
    """
    Generates a 3D point within dim boundaries
    Args:
        dim (int): Boundaries of the coordinates

    Returns:
        point (np.ndarray): Generated point
    """
    point = np.random.rand(3) * dim
    return point


def monte_carlo(dim, hull, number):
    """
    Monte Carlo volume estimation algorithm

    Args:
        dim (int): Dimensions of the box
        hull (hull object): Convex hull of the structure
        number (int): Number of points to generate

    Returns:
        ratio (float): Ratio of number of points generated inside the volume and overall number of points
    """
    in_volume = 0
    out_volume = 0

    for _ in range(number):
        point = generate_point(dim)

        if _is_inside(point, hull):
            in_volume += 1
        else:
            out_volume += 1

    ratio = in_volume / (out_volume + in_volume)
    return ratio


def monte_carlo_volume(dim, hull, number):
    """
    Utility function responsible for rescaling and calling the actual algorithm

    Args:
        dim (int): Dimensions of the box
        hull (hull object): Hull object
        number (int): Number of points to generate
        rescale (int): Rescale factor

    Returns:
        float: Estimated volume of the structure
    """

    pbc_vol = dim ** 3
    pbc_sys_ratio = monte_carlo(dim, hull, number=number)

    return pbc_sys_ratio * pbc_vol  # , points



import abc


class MoleculeSystem(abc.ABC):
    """ Every PUCHIK class should derive from this in the future. """
    @abc.abstractmethod
    def __init__(self, traj: str, top=None):
        ...

    @abc.abstractmethod
    def select_atoms(self, selection: str) -> None:
        ...



cimport cython
import numpy as np
cimport numpy as np
from pygel3d import hmesh
from libc.math cimport fabs

# from scipy.spatial import ConvexHull
np.import_array()


def find_distance(hull, np.ndarray points):
    cdef np.ndarray p, res
    cdef float d
    cdef int i, p_length
    # Construct PyGEL Manifold from the convex hull
    m = hmesh.Manifold()
    for s in hull.simplices:
        m.add_face(hull.points[s])

    dist = hmesh.MeshDistance(m)
    p_length = points.shape[0]
    res = np.zeros(p_length)

    for i in range(p_length):
        p = points[i]
        # Get the distance to the point
        # But don't trust its sign, because of possible
        # wrong orientation of mesh face
        d = dist.signed_distance(p)

        # Correct the sign with ray inside test
        if dist.ray_inside_test(p):
            if d > 0:
                d *= -1
        else:
            if d < 0:
                d *= -1

        res[i] = d

    return res


def _is_inside(np.ndarray point, hull, alpha_shape=False) -> bool:
    if alpha_shape:
        vertices = hull.points
        cells = hull.cells
        return point_in_alpha_shape(point, cells, vertices)
    return point_in_hull(point, hull)


def point_in_hull(np.ndarray point, hull):
    cdef double tolerance
    tolerance = 1e-12

    return all(
        (np.dot(eq[:-1], point) + eq[-1] <= tolerance)
        for eq in hull.equations)


def point_in_alpha_shape(np.ndarray point, np.ndarray cells, np.ndarray vertices) -> bool:
    cdef np.ndarray cell, tetrahedron_points

    for cell in cells:
        tetrahedron_points = vertices[cell]
        if point_in_tetrahedron(point, tetrahedron_points):
            return True
    return False

def point_in_tetrahedron(np.ndarray point, np.ndarray tetrahedron_points) -> bool:
    cdef np.ndarray v0, v1, v2, v_point
    cdef float denom, u, v, w, t

    v0 = tetrahedron_points[1] - tetrahedron_points[0]
    v1 = tetrahedron_points[2] - tetrahedron_points[0]
    v2 = tetrahedron_points[3] - tetrahedron_points[0]
    v_point = point - tetrahedron_points[0]

    # Calculate determinants
    denom = np.linalg.det(np.column_stack((v0, v1, v2)))
    if denom == 0:
        return False

    # Calculate barycentric coordinates
    u = np.linalg.det(np.column_stack((v_point, v1, v2))) / denom
    v = np.linalg.det(np.column_stack((v0, v_point, v2))) / denom
    w = np.linalg.det(np.column_stack((v0, v1, v_point))) / denom
    t = 1 - u - v - w

    # Check if the point is inside the tetrahedron
    return (u >= 0) and (v >= 0) and (w >= 0) and (t >= 0)


import pytest
import os
from numpy import isclose
from PUCHIK.grid_project.core.interface import Interface

TEST_DIR = './PUCHIK/test/test_structures'


def test_object_creation():
    m = Interface(
        os.path.join(TEST_DIR, 'InP_cylinder.pdb')
    )


def test_create_mesh():
    m = Interface(
        os.path.join(TEST_DIR, 'InP_cylinder.pdb')
    )

    m.calculate_mesh('resname UNL')


def test_create_hull():
    m = Interface(
        os.path.join(TEST_DIR, 'InP_cylinder.pdb')
    )
    m.select_structure('resname UNL')
    m._create_hull()


def test_calculate_volume():
    m = Interface(
        os.path.join(TEST_DIR, 'InP_cylinder.pdb')
    )
    m.select_structure('resname UNL')
    v = m.calculate_volume()
    assert isclose(v, 146450.0), f'Volume should be close to {146450.0}'


def test_create_alpha_hull():
    m = Interface(
        os.path.join(TEST_DIR, 'InP_cylinder.pdb')
    )
    m.use_alpha_shape = True
    m.select_structure('resname UNL')
    m._create_hull()



import subprocess
import numpy as np
from dotenv import load_dotenv
import os
from pathlib import Path

# Load .env to get alpha_shape exe
load_dotenv()


class AlphaShape:
    """ Python wrapper for AlphaShaper.exe """
    def __init__(self, points):
        self._points = points  # All points
        self._cells = None
        self._simplices = None

    @property
    def cells(self):
        return self._points

    @cells.setter
    def cells(self, new_cells):
        self._cells = new_cells

    @property
    def points(self):
        return self._points

    @points.setter
    def points(self, new_points):
        self._simplices = new_points

    @property
    def simplices(self):
        return self._simplices

    @simplices.setter
    def simplices(self, new_simplices):
        self._simplices = new_simplices

    def calculate_as(self, frame_num, alpha=-1):
        """
        Calculate the alpha shape for frame <frame_num>
        :param frame_num:
        :param alpha:
        :return:
        """
        temp_file_name = f'./.temp_frame_{frame_num}.xyz'
        temp_output_facets_file_name = f'output_facets_{frame_num}.txt'
        temp_output_cells_file_name = f'output_cells_{frame_num}.txt'

        output_file_suffix = f'{frame_num}'
        alpha_shaper_exe = Path(__file__).resolve().parent.parent / 'alpha_shaper' / 'AlphaShaper.exe'
        if not os.path.exists(alpha_shaper_exe):
            print(alpha_shaper_exe)
            raise FileNotFoundError("Couldn't find the executable.")

        np.savetxt(temp_file_name, self.points, header=f'{len(self.points)}', comments='')
        proc = subprocess.run([alpha_shaper_exe, temp_file_name, f'{alpha}', output_file_suffix],
                              capture_output=True, text=True)

        self.simplices = np.loadtxt(temp_output_facets_file_name, dtype=int)
        self.cells = np.loadtxt(temp_output_cells_file_name, dtype=int)

        os.remove(temp_file_name)
        os.remove(temp_output_facets_file_name)
        os.remove(temp_output_cells_file_name)

        return self



import numpy as np
import MDAnalysis as mda
from MDAnalysis.lib.distances import distance_array
from MDAnalysis.transformations import unwrap

from PUCHIK.grid_project.utilities.MoleculeSystem import MoleculeSystem

CLUSTER_SEARCH_CUTOFF = 4.7


class ClusterSearch(MoleculeSystem):
    """ A class to search for clusters. The algorithm is somewhat similar to a depth-first search algorithm.
    Select atoms to cluster using select_atoms method and then run find_clusters to return the list of ids in each
    cluster for each frame. !TODO Currently we consider isolated atoms as separate clusters and might need to change it
    """
    def __init__(self, top_path: str, trj_path=None):
        self.ag = None
        if not trj_path:
            self.u = mda.Universe(top_path)
        else:
            self.u = mda.Universe(top_path, trj_path)

    def select_atoms(self, selection: str) -> None:
        self.ag = self.u.select_atoms(selection)

    def _dfs(self, index: int, positions: np.ndarray, cluster: list, checked: list) -> list:
        checked.append(index)

        for i in range(len(positions)):

            if i != index and i not in checked:
                distances = distance_array(positions[index], positions[i])

                if (distances < CLUSTER_SEARCH_CUTOFF).any():
                    cluster.append(i)
                    self._dfs(i, positions, cluster, checked)

        return cluster

    def find_clusters(self):
        n_res = self.ag.n_residues
        resids = np.unique(self.ag.resids)
        n_atoms = self.ag.n_atoms // n_res
        clusters = []

        positions = np.zeros((len(self.u.trajectory), n_res, n_atoms, 3))
        for ts in self.u.trajectory:
            current_frame_clusters = []
            checked_residues = []

            for i in range(n_res):
                for j in range(n_atoms):
                    positions[ts.frame, i, j] = self.ag[i * n_atoms + j].position

            for i, position in enumerate(positions[ts.frame]):
                if i not in checked_residues:
                    current_frame_clusters.append(resids[self._dfs(i, positions[ts.frame], [i], checked_residues)])
            clusters.append(current_frame_clusters)
        return clusters


def center_to_file(u, selection, o_filename=None, ext=None):
    """
    A utility function to center the system around an atom.

    :param u: MDAnalysis Universe object
    :param selection: Selection of the atom to be centered
    :param o_filename: Output trajectory file name
    :param ext: Extension of the output trajectory file
    :return:
    """

    # !TODO Implement trajectory slicing here
    all_atoms = u.select_atoms('all')

    with mda.Writer(f'{o_filename}.xtc', all_atoms.n_atoms) as w:
        for _ in u.trajectory:
            pbc_dim = u.dimensions[0]
            all_atom_pos = all_atoms.positions
            center_atom_pos = u.select_atoms(selection).positions
            new_pos = _translate_system(all_atom_pos, center_atom_pos, pbc_dim)
            all_atoms.positions = new_pos

            w.write(all_atoms)


def center_in_memory(u, selection):
    """
    A utility function to center the system around an atom inplace.

    :param u: MDAnalysis Universe object
    :param selection: Selection of the atom to be centered
    :return:
    """
    u.transfer_to_memory()

    all_atoms = u.select_atoms('all')

    for ts in u.trajectory:
        pbc_dim = u.dimensions[0]
        all_atom_pos = all_atoms.positions
        center_atom_pos = u.select_atoms(selection).positions

        new_pos = _translate_system(all_atom_pos, center_atom_pos, pbc_dim)

        ts.positions = new_pos


def _translate_system(positions: np.ndarray, center_atom_pos: np.ndarray, pbc_dim: float) -> np.ndarray:
    """
    A utility function to translate the system towards the center and put all atoms back in the box.

    :param positions: Positions to translate
    :param center_atom_pos: Positions of the new center
    :param pbc_dim: Dimensions of PBC
    :return: New positions
    """
    pbc_center = np.array((pbc_dim,) * 3) / 2

    # Translate everything to the desired position
    translation_vec = center_atom_pos - pbc_center
    new_pos = positions - translation_vec

    # Bring atoms back into the PBC
    new_pos = np.mod(new_pos, pbc_dim)

    # u.select_atoms('all').positions = new_pos
    return new_pos



from multiprocessing import cpu_count


DEBUG = False  # Change this to True to see logs
CPU_COUNT = cpu_count()
TQDM_BAR_FORMAT = '\033[37m{percentage:3.0f}%|{bar:30}\033[37m|[Estimated time remaining: {remaining}]\033[0m'






import os
from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import numpy

# Import the current version number
# from PUCHIK._version import __version__
# import tomllib  # no tomllib before 3.11


def get_current_version():
    return "1.2.4"
#     with open("pyproject.toml", "rb") as f:
#         data = tomllib.load(f)
#
#     return data['project']['version']


extensions = [
    Extension(
        name='PUCHIK.grid_project.core.utils',
        sources=['PUCHIK/grid_project/core/utils.pyx'],
        include_dirs=[numpy.get_include(), 'PUCHIK/grid_project/core'],
    )
]

setup(
    name='PUCHIK',
    version=get_current_version(),
    description='Python Utility for Characterizing Heterogeneous Interfaces and Kinetics',
    url='https://github.com/hrachishkhanyan/grid_project',
    author='H. Ishkhanyan',
    author_email='hrachya.ishkhanyan@kcl.ac.uk',
    license='MIT',
    provides=['PUCHIK'],
    packages=['PUCHIK'],
    ext_modules=cythonize(extensions),
)
